#include "nes.h"
#include "map.h"

#define BG_NAMTBL_W (SCREEN_W/8)
#define BG_NAMTBL_H (SCREEN_H/8)
#define BG_ATTR_W   8
#define BG_ATTR_H   8

#define MAP_FIELD_W 12
#define MAP_FIELD_H 11

#define MAP_SYNBOL(chr, mcp) (((chr + 1) << 4)|(mcp))
#define GET_CHAR_ID(m)       ((((m) >> 4) & 0x0f) -1)
#define GET_CHIP_ID(m)       ((m) & 0x0f)

// }bv`bv
const unsigned char mapchips[][4]  = {
	{0x00,0x00,0x00,0x00}, // Ƃ
	{0x18,0x19,0x1a,0x1b}, // ]{[
	{0x1c,0x1d,0x1e,0x1f}, // S[̊
	{0x14,0x15,0x16,0x17}, // 󂹂Ȃ
	{0x21,0x21,0x21,0x21}, // 󂹂ubN
	{0x21,0x21,0x21,0x21}, // ubN
};

// BGe[u
const unsigned char attrbg[] = {
	0x02, // Ƃ
	0x02, // ]{[
	0x02, // S[̊
	0x00, // 󂹂Ȃ
	0x01, // 󂹂ubN
	0x00, // ubN
};

// BGppbg
const unsigned char palettebg[] = {
	0x30, 0x31, 0x21, 0x11,
	0x30, 0x26, 0x16, 0x05,
	0x30, 0x34, 0x25, 0x15,
	0x0f, 0x31, 0x10, 0x00,
};

// }bve[u
extern const unsigned short maps[];

// Q[̃}bv
unsigned char gamemap[MAP_FIELD_H][MAP_FIELD_W];

void map_init(char scene) {

	// pbgݒ
	SetPalette(palettebg, 0);

	// wiݒ
	FillNameTbl( 0,  0, 32, 30, 0x03);
	FillNameTbl( 3,  3, 26, 24, 0x00);
	FillNameTbl( 3,  2, 26,  1, 0x0a);
	FillNameTbl( 3, 27, 26,  1, 0x0b);
	FillNameTbl( 2,  3,  1, 24, 0x09);
	FillNameTbl(29,  3,  1, 24, 0x08);

	// wiݒ
	FillAttribute(BG1_ATTR_HADR, BG1_ATTR_LADR, 0, 64);

	// }bv̏
#if 0
	{
		unsigned char x, y;

		for (y = 0; y < MAP_FIELD_H; ++y) {
			for (x = 0; x < MAP_FIELD_W; ++x) {
				gamemap[y][x] = y;
			}
		}
	}
#endif
	{
		unsigned char* map = (char *)maps[scene];
		unsigned char x, y;
		unsigned char cnt = 0;

		for (y = 0; y < MAP_FIELD_H; ++y) {
			for (x = 0; x < MAP_FIELD_W; ++x) {
				unsigned char m = map[cnt++];
				//unsigned char id;

				// }bv̎擾
				gamemap[y][x] = m;

				// TODO : LN^̎擾
				//id = GET_CHAR_ID(m);
			}
		}
	}

	// }bv񂩂wiݒ
	{
		unsigned char nam[2][MAP_FIELD_W*2];
		unsigned char x, y;

		for (y = 0; y < MAP_FIELD_H; ++y) {
			for (x = 0; x < MAP_FIELD_W; ++x) {
				unsigned char m = gamemap[y][x];
				const unsigned char *mc = mapchips[GET_CHIP_ID(m)];
				unsigned char dx = x * 2;

				nam[0][dx + 0] = mc[0];
				nam[0][dx + 1] = mc[1];
				nam[1][dx + 0] = mc[2];
				nam[1][dx + 1] = mc[3];
			}

			SetNameTbl(nam[0], (BG_NAMTBL_W-MAP_FIELD_W*2)/2, (BG_NAMTBL_H-MAP_FIELD_H*2)/2+y*2  , MAP_FIELD_W*2, 1);
			SetNameTbl(nam[1], (BG_NAMTBL_W-MAP_FIELD_W*2)/2, (BG_NAMTBL_H-MAP_FIELD_H*2)/2+y*2+1, MAP_FIELD_W*2, 1);
		}
	}

	// }bv񂩂瑮ݒ
	{
		unsigned char attrtbl[BG_ATTR_H][BG_ATTR_W];
		unsigned char x, y;

		for (y = 0; y < BG_ATTR_H; ++y) {
			for (x = 0; x < BG_ATTR_W; ++x) {
				attrtbl[y][x] = 0;
			}
		}

		for (y = 0; y < MAP_FIELD_H; ++y) {
			unsigned char ay = y / 2 + (BG_NAMTBL_H-MAP_FIELD_H*2)/8;

			for (x = 0; x < MAP_FIELD_W; ++x) {
				unsigned char m = GET_CHIP_ID(gamemap[y][x]);
				unsigned char attr = attrbg[m];
				unsigned char ax = x / 2 + (BG_NAMTBL_W-MAP_FIELD_W*2)/8;
				unsigned char shift;

				shift  = (x & 1)? 2: 0;
				if (y & 1) shift += 4;

				attrtbl[ay][ax] |= attr << shift;
			}
		}

		for (y = 0; y < BG_ATTR_H; ++y) {
			unsigned char ladr = BG1_ATTR_LADR + y * BG_ATTR_W;
			SetAttribute(BG1_ATTR_HADR, ladr, attrtbl[y], sizeof(attrtbl[0]));
		}
	}
}

// V{`
#define _   MCP_FIELD
#define O   MCP_BALL
#define G   MCP_GOAL
#define W   MCP_WALL
#define B   MCP_BLOCK
#define F   MCP_FBLOCK

// }bv
const unsigned char map01[] = {
	 _, _, _, _, _, _, _, _, _, _, _, _,
	 _, _, _, _, _, _, _, _, _, _, _, _,
	 _, _, _, _, _, _, _, _, _, O, _, _,
	 _, _, _, _, B, _, _, _, O, _, O, _,
	 _, _, _, B, _, B, _, _, _, O, _, _,
	 _, _, B, _, _, _, B, _, _, _, _, _,
	 _, _, _, _, _, _, _, B, _, _, _, _,
	 _, _, _, _, _, _, _, _, B, _, _, _,
	 _, _, _, _, _, _, _, _, _, B, _, _,
	 _, _, _, _, _, _, _, _, _, _, _, _,
	 _, _, _, _, _, _, _, _, _, _, _, _,
};

const unsigned char map02[] = {
	 _, _, _, _, _, _, _, _, _, _, _, _,
	 _, _, _, _, F, _, _, _, _, _, _, _,
	 _, _, _, _, F, _, _, _, _, _, _, _,
	 _, _, F, F, F, F, F, _, _, _, _, _,
	 _, _, _, _, F, _, _, _, _, _, _, _,
	 _, _, _, _, F, F, F, _, _, _, _, _,
	 _, _, _, _, F, _, _, F, _, _, _, _,
	 _, _, _, _, _, _, _, F, _, _, _, _,
	 _, _, _, _, _, _, _, F, _, _, _, _,
	 _, _, _, _, _, F, F, _, _, _, _, _,
	 _, _, _, _, _, _, _, _, _, _, _, _,
};

const unsigned char map03[] = {
	 _, _, _, _, _, _, _, _, _, _, _, _,
	 _, _, _, _, _, _, W, _, _, _, W, _,
	 _, W, _, _, _, _, W, _, _, W, _, W,
	 _, W, _, W, W, W, W, W, _, _, W, _,
	 _, W, _, _, _, _, W, _, _, _, _, _,
	 _, W, _, _, _, _, W, _, _, _, _, _,
	 _, _, W, _, _, W, W, W, _, _, _, _,
	 _, _, W, _, W, _, W, _, _, _, _, _,
	 _, _, W, _, W, _, W, _, _, _, _, _,
	 _, _, W, _, _, W, _, _, _, _, _, _,
	 _, _, _, _, _, _, _, _, _, _, _, _,
};

const unsigned char map04[] = {
	 _, _, _, _, _, _, _, _, _, _, _, _,
	 _, _, _, _, B, _, _, _, _, _, _, _,
	 _, _, _, _, B, _, _, _, _, _, _, _,
	 _, _, B, B, B, B, _, B, _, _, _, _,
	 _, _, _, B, _, _, _, _, B, _, _, _,
	 _, _, _, B, _, _, B, _, _, _, _, _,
	 _, _, B, _, _, _, B, _, _, _, _, _,
	 _, _, B, _, _, B, B, B, _, _, _, _,
	 _, _, _, _, B, _, B, _, _, _, _, _,
	 _, _, _, _, _, B, _, _, _, _, _, _,
	 _, _, _, _, _, _, _, _, _, _, _, _,
};

const unsigned char map05[] = {
	 _, _, _, _, _, _, _, _, _, _, _, _,
	 _, F, F, F, F, F, F, F, F, F, _, _,
	 _, _, _, _, _, F, _, _, _, _, _, _,
	 _, _, _, _, _, F, _, _, _, _, _, _,
	 _, _, F, F, F, F, F, F, F, _, _, _,
	 _, _, _, _, F, _, F, _, _, _, _, _,
	 _, _, _, _, F, _, F, _, _, _, _, _,
	 _, _, _, F, _, _, _, F, _, _, _, _,
	 _, _, _, F, _, _, _, F, _, _, _, _,
	 _, _, F, _, _, _, _, _, F, _, _, _,
	 _, F, _, _, _, _, _, _, _, F, _, _,
	 _, _, _, _, _, _, _, _, _, _, _, _,
};

// }bve[u
const unsigned short maps[] = {
	(unsigned short)map01,
	(unsigned short)map02,
	(unsigned short)map03,
	(unsigned short)map04,
	(unsigned short)map05,
};
