#include "nes.h"
#include "nsd.h"
#include "math.h"
#include "sound.h"
#include "chara.h"
#include "map.h"

#define SCENE_MAX 5
#define SCENE_START_WAIT 120
#define SCENE_GOAL_WAIT  120

#define STATUS_BUSY 1
#define STATUS_GOAL 2

enum {
	SCENE_STATUS_START = 0,
	SCENE_STATUS_PLAY,
	SCENE_STATUS_GOAL,
	SCENE_STATUS_CLEAR,
	SCENE_STATUS_OVER,
	SCENE_STATUS_PAUSE = 0x80,
};

const char forward_offset1[][2] = {
	{ 0, -5},
	{ 5,  0},
	{ 0,  5},
	{-5,  0},
};

const char forward_offset2[][2] = {
	{  0, -21},
	{ 21,   0},
	{  0,  21},
	{-21,   0},
};

// C
void NesMain()
{
	unsigned char frame;
	unsigned char scene;
	unsigned char status;
	Chara chara[10];
	Ball ball;

	// ϐ
	frame = 0;
	scene = 0;
	status = SCENE_STATUS_START;
	FillMemory(&ball, 0, sizeof(ball));

	// PPȔ
	InitPPU();

	// hCo
	nsd_init();

	// DMA
	InitDMA();

	// BG]R}ḧ揉
	InitCMD();

	// 
	srand(14, 10, 11);

	// LN^
	chara_init(chara, NUMBEROF(chara));

	// }bv
	map_init(scene, chara, NUMBEROF(chara));

	// wi̕\ʒuݒ
	SetScroll(0, 0);

	// `Jn(NMI荞ݗL)
	SetPPU(0x88, 0x1e);

	while (1) {
		// ݏI҂
		Sync();

		// BG]R}h
		ResetCMD();

		// DMA]\񏉊
		ResetDMA();

		// Rg[mF
		CheckPad();

		// Q[Ŝ̐
		switch (status) {
		default:
		case SCENE_STATUS_START:
			if (frame >= SCENE_START_WAIT) {
				status = SCENE_STATUS_PLAY;

				// tJn
				nsd_play_bgm(BGM_PENGUIN);
			}
			break;

		case SCENE_STATUS_PLAY:
			// |[Y
			if (ButtonDown(0, BTN_START)) {
				status |= SCENE_STATUS_PAUSE;

				// tꎞ~
				nsd_pause_bgm();
			}
			break;

		case SCENE_STATUS_PLAY|SCENE_STATUS_PAUSE:
			// |[Y
			if (ButtonDown(0, BTN_START)) {
				status &= ~SCENE_STATUS_PAUSE;

				// tĊJ
				nsd_resume_bgm();
			}
			break;

		case SCENE_STATUS_GOAL:
			if (frame >= SCENE_GOAL_WAIT) {
				status = SCENE_STATUS_CLEAR;
			}
			break;

		case SCENE_STATUS_CLEAR:
			// ϐ
			scene = (scene + 1) % SCENE_MAX;

		case SCENE_STATUS_OVER:
			// t~
			nsd_stop_bgm();

			// ϐ
			status = SCENE_STATUS_START;
			frame = 0;
			FillMemory(&ball, 0, sizeof(ball));

			// PPȔ
			InitPPU();

			// DMA
			InitDMA();

			// LN^
			chara_init(chara, NUMBEROF(chara));

			// }bv
			map_init(scene, chara, NUMBEROF(chara));

			// wi̕\ʒuݒ
			SetScroll(0, 0);

			// `Jn(NMI荞ݗL)
			SetPPU(0x88, 0x1e);

			continue;
		}

		if (!(status & SCENE_STATUS_PAUSE)) {

			if (status == SCENE_STATUS_PLAY) {
				unsigned char i;

				// {[̈ړ
				if (ball.status) {
					nsd_play_se(SE_MOVE);
					map_move(ball.x, ball.y, ball.dir);

					// S[
					if (ball.status == STATUS_GOAL) {
						status = SCENE_STATUS_GOAL;
						frame = 0;
					}
					ball.status = 0;
				}

				// vC[L̐
				for (i = 0; i < 2; ++i) {
					Chara *p = &chara[i];
					unsigned char x = p->x;
					unsigned char y = p->y;

					char vx = 0;
					char vy = 0;

					// L[͂ɂړ
					if (ButtonPress(i, BTN_RIGHT)) { vx++; p->dir = 1; }
					if (ButtonPress(i, BTN_LEFT )) { vx--; p->dir = 3; }
					if (ButtonPress(i, BTN_UP   )) { vy--; p->dir = 0; }
					if (ButtonPress(i, BTN_DOWN )) { vy++; p->dir = 2; }

					// Oɂړ
					vx += p->dx; p->dx = 0;
					vy += p->dy; p->dy = 0;

					// n`Ŕ肵Ă݂
					if (vx || vy) {
						// LN^ʒũ}bv擾
						unsigned char mcp1;
						unsigned char x = p->x + vx;
						unsigned char y = p->y + vy;

						if (vx & 0x80) x -= 4;
						else if (vx)   x += 4;

						if (vy & 0x80) y -= 4;
						else if (vy)   y += 4;

						mcp1 = map_getmcp(x, y);

						if (mcp1 == MCP_FIELD || mcp1 == MAP_SYNBOL(i,MCP_FIELD) || GET_CHIP_ID(mcp1) == MCP_GOAL) {
							unsigned char mcp0 = map_getmcp(p->x, p->y);

							map_setmcp(p->x, p->y, REMOVE_CHAR(mcp0));
							map_setmcp(x, y, MAP_SYNBOL(i,MCP_FIELD));

							p->x += vx;
							p->y += vy;
						}
					}
					else if (ButtonDown(i, BTN_A)) {
						// is̃}bv擾
						unsigned char x1, y1;
						unsigned char mcp1;

						x1 = p->x + forward_offset1[p->dir][0];
						y1 = p->y + forward_offset1[p->dir][1];
						mcp1 = map_getmcp(x1, y1);

						switch (mcp1) {
						case MCP_BLOCK:
							nsd_play_se(SE_BREAK);
							map_set(x1, y1, MCP_FIELD);
							break;

						case MCP_BALL:
							{
								unsigned char x2, y2;
								unsigned char mcp2;

								// {[̐mF
								x2 = p->x + forward_offset2[p->dir][0];
								y2 = p->y + forward_offset2[p->dir][1];
								mcp2 = map_getmcp(x2, y2);

								if (mcp2 == MCP_GOAL) ball.status = STATUS_GOAL;
								else if (mcp2 != MCP_FIELD) break;
								else ball.status = STATUS_BUSY;

								nsd_play_se(SE_MOVE);
								map_move(x1, y1, p->dir);
								ball.x = x1;
								ball.y = y1;
								ball.dir = p->dir;
							}
							break;
						}
					}
				}

				// GL̐
				for (i = 2; i < NUMBEROF(chara); ++i) {
					Chara* e = &chara[i];

					switch (e->chr) {
					case CHR_SLIME:
						break;
					case CHR_WIZARD:
						break;
					case CHR_FIRE_BALL:
						break;
					}
				}
			}

			// t[JE^XV
			++frame;
		}

		// LN^̕\
		chara_draw(chara, NUMBEROF(chara), (frame >> 4) & 3);

		// BG]f[^ݒI
		SetEndCMD();
	}
}

//NMI荞
void NMIProc(void)
{
	// ҂JnmF
	if (CheckSync()) {

		// DMA]
		SendDMA();

		// BG]R}hs
		SendCMD();

		// XN[l̍Đݒ
		ResetScroll();

		// TEhhCoĂяo
		nsd_main();

		// ݒ
		DoneSync();
	}
}
