#include "nes.h"
#include "nsd.h"
#include "math.h"
#include "sound.h"
#include "chara.h"
#include "map.h"

// XvCgppbg
const char palettesp[] = {
	0x30, 0x26, 0x30, 0x01,
	0x2c, 0x16, 0x30, 0x25,
	0x0f, 0x29, 0x39, 0x1a,
	0x30, 0x3d, 0x1d, 0x00,
};

// LN^e[u
const char charatable[][4] = {
	{0x01,0x02,0x1f,0x20}, // 
	{0x03,0x04,0x21,0x22}, // ʑOi
	{0x05,0x06,0x23,0x24}, // E
	{0x07,0x08,0x25,0x26}, // EOi
	{0x09,0x0a,0x27,0x28}, // w
	{0x0b,0x0c,0x29,0x2a}, // wʑOi
};

// Aj[Ve[u
const char animetable[][4][2] = {
	{{4, 0x00}, {5,0x00}, {4, 0x00}, {5, 0x40}}, // 
	{{2, 0x40}, {3,0x40}, {2, 0x40}, {3, 0x40}}, // 
	{{0, 0x00}, {1,0x00}, {0, 0x00}, {1, 0x40}}, // 
	{{2, 0x00}, {3,0x00}, {2, 0x00}, {3, 0x00}}, // E
};

// XCLN^e[u
const char slime_chrtbl[][4] = {
	{0x0d,0x0e,0x2b,0x2c}, // 
	{0x0f,0x10,0x2d,0x2e}, // 
	{0x11,0x12,0x2f,0x30}, // 
};

// XCAj[Ve[u
const char slime_anmtbl[][4][2] = {
	{{1, 2}, {0, 2}, {1, 2}, {2, 2}}, // 
	{{1, 2}, {2, 2}, {1, 2}, {0, 2}}, // 
	{{1, 2}, {0, 2}, {1, 2}, {2, 2}}, // 
	{{1, 2}, {2, 2}, {1, 2}, {0, 2}}, // E
};

// @gLN^e[u
const char wizard_chrtbl[][4] = {
	{0x13,0x14,0x31,0x32}, // 
	{0x15,0x16,0x33,0x34}, // ʑOi
	{0x17,0x18,0x35,0x36}, // E
	{0x19,0x1a,0x37,0x38}, // EOi
	{0x1b,0x1c,0x39,0x3a}, // w
	{0x1d,0x1e,0x3b,0x3c}, // wʑOi
};

// LN^pbgwe[u
const unsigned char charapal[] = {
	0x00,
	0x01,
	0x02,
	0x03,
};

// ΋ˏo
const char firevector[][2] = {
	{ 0,-4},
	{ 1,-3},
	{ 3,-3},
	{ 3,-1},
	{ 4, 0},
	{ 3, 1},
	{ 3, 3},
	{ 1, 3},
	{ 0, 4},
	{-1, 3},
	{-3, 3},
	{-3, 1},
	{-4, 0},
	{-3,-3},
	{-3,-1},
};

Chara chara[CHARA_MAX];
Ball ball;

void __fastcall__ chara_init(void) {

	unsigned char i;

	// pbgݒ
	SetPalette(palettesp, 1);

	// LN^̏
	FillMemory(chara, 0, sizeof(chara));

	for (i = 0; i < CHARA_MAX; ++i) {
		Chara* p = &chara[i];

		p->chr = (i < 2)? i: CHR_NOTHING;
		p->dir = 2;
	}
}

void __fastcall__ chara_draw(unsigned char frame) {
	unsigned char i;

	for (i = 0; i < CHARA_MAX; ++i) {
		Chara *p = &chara[i];

		switch (p->chr) {
		case CHR_BLUE_PENGUIN:
		case CHR_RED_PENGUIN:
		case CHR_WIZARD:
			{
				const char* anime = animetable[p->dir][frame];
				const char* chara = (p->chr == CHR_WIZARD)?
					wizard_chrtbl[anime[0]]: charatable[anime[0]];
				unsigned char flag = anime[1] | charapal[p->chr];

				if (flag & 0x40) {
					SetDMA(p->x-8, p->y-12, chara[1], flag);
					SetDMA(p->x  , p->y-12, chara[0], flag);
					SetDMA(p->x-8, p->y- 4, chara[3], flag);
					SetDMA(p->x  , p->y- 4, chara[2], flag);

				} else {
					SetDMA(p->x-8, p->y-12, chara[0], flag);
					SetDMA(p->x  , p->y-12, chara[1], flag);
					SetDMA(p->x-8, p->y- 4, chara[2], flag);
					SetDMA(p->x  , p->y- 4, chara[3], flag);
				}
				break;
			}

		case CHR_SLIME:
			{
				const char *anime = slime_anmtbl[p->dir][frame];
				const char *chara = slime_chrtbl[anime[0]];
				unsigned char flag = anime[1];

				SetDMA(p->x-8, p->y-12, chara[0], flag);
				SetDMA(p->x  , p->y-12, chara[1], flag);
				SetDMA(p->x-8, p->y- 4, chara[2], flag);
				SetDMA(p->x  , p->y- 4, chara[3], flag);
				break;
			}

		case CHR_FIRE_BALL:
			SetDMA(p->x-4, p->y-4, 0x3d, 1);
			break;
		}
	}
}



#define SLIME_ACTION_MASK 0x3f

enum {
	SLIME_ACTION_INIT =  0,
	SLIME_ACTION_DIR,
	SLIME_ACTION_MOVE,
	SLIME_ACTION_STOP,
	SLIME_ACTION_MAX
};

unsigned char __fastcall__  action_slime(Chara* s) {
	unsigned char action = s->status & SLIME_ACTION_MASK;
	unsigned char i;

	switch (action) {
	case SLIME_ACTION_INIT:
		action = SLIME_ACTION_DIR | STATUS_BUSY;
		break;

	case SLIME_ACTION_DIR:
		// vC[ɋ߂Â
		{
			unsigned char x0 = s->x >> 4;
			unsigned char y0 = s->y >> 4;
			unsigned char x1, y1;

			for (i = 0; i < 2; ++i) {
				x1 = chara[i].x >> 4;
				y1 = chara[i].y >> 4;

				if (x0 == x1 && y0 != y1) {
					s->dx = 0;
					s->dy = (y0 < y1)? 1: -1;
					action = SLIME_ACTION_MOVE;
					break;
				}

				if (y0 == y1 && x0 != x1) {
					s->dx = (x0 < x1)? 1: -1;
					s->dy = 0;
					action = SLIME_ACTION_MOVE;
					break;
				}
			}

			if (action != SLIME_ACTION_MOVE) {
				rand(i);
				if (i > 127) action = SLIME_ACTION_STOP;
			}
		}
		break;

	case SLIME_ACTION_MOVE:
		{
			unsigned char x = s->x + s->dx;
			unsigned char y = s->y + s->dy;
			unsigned char mcp1;

			if (s->dx & 0x80) x -= 4;
			else if (s->dx)   x += 4;

			if (s->dy & 0x80) y -= 4;
			else if (s->dy)   y += 4;

			mcp1 = map_getmcp(x, y);

			if (mcp1 == MCP_FIELD || mcp1 == MAP_SYNBOL(s->id,MCP_FIELD)) {
				unsigned char mcp0 = map_getmcp(s->x, s->y);

				map_setmcp(s->x, s->y, REMOVE_CHAR(mcp0));
				map_setmcp(x, y, MAP_SYNBOL(s->id,MCP_FIELD));

				s->x += s->dx;
				s->y += s->dy;
			}

			rand(i);
			if (i < 32) {
				action = SLIME_ACTION_STOP;
			}
		}
		break;

	case SLIME_ACTION_STOP:
		s->status &= ~STATUS_BUSY;
		action = SLIME_ACTION_INIT;
		break;
	}

	s->status = (s->status & ~SLIME_ACTION_MASK) | action;

	return s->status;
}


#define WIZARD_ACTION_MASK 0x3f

enum {
	WIZARD_ACTION_INIT =  0,
	WIZARD_ACTION_WAIT,
	WIZARD_ACTION_FIRE,
	WIZARD_ACTION_STOP,
	WIZARD_ACTION_MAX
};

unsigned char __fastcall__  action_wizard(Chara* w) {
	unsigned char action = w->status & WIZARD_ACTION_MASK;
	unsigned char i;

	switch (action) {
	case WIZARD_ACTION_INIT:
		action = WIZARD_ACTION_WAIT | STATUS_BUSY;
		break;

	case WIZARD_ACTION_WAIT:
		// vC[ɕ
		w->dir = (chara[0].y >= w->y)? 2: 0;

		// _ɔ˃V[PXֈڍs
		rand(i);
		if (i & 0x10) action = WIZARD_ACTION_FIRE;
		break;

	case WIZARD_ACTION_FIRE:
		// t@C[{[I
		{
			Chara* fireball;

			action = WIZARD_ACTION_STOP;

			// Ǘ̈͂QȂ̂ŗ͋Z
			if (chara[FIRE_BALL_TOP].chr == CHR_NOTHING) {
				fireball = &chara[FIRE_BALL_TOP];
			} else if (chara[FIRE_BALL_TOP+1].chr == CHR_NOTHING) {
				fireball = &chara[FIRE_BALL_TOP+1];
			} else {
				break;
			}

			// ˏo߂
			rand(i);
			i = ((w->dir << 2) + ((i >> 4) & 7) - 4) & 0x0f;

			// ˏo
			fireball->chr = CHR_FIRE_BALL;
			fireball->x = w->x;
			fireball->y = w->y;
			fireball->dx = firevector[i][0];
			fireball->dy = firevector[i][1];
		}
		break;

	case WIZARD_ACTION_STOP:
		w->status &= ~STATUS_BUSY;
		action = WIZARD_ACTION_INIT;
		break;
	}

	w->status = (w->status & ~WIZARD_ACTION_MASK) | action;

	return w->status;
}


#define FIREBALL_ACTION_MASK 0x3f

enum {
	FIREBALL_ACTION_INIT =  0,
	FIREBALL_ACTION_MOVE,
	FIREBALL_ACTION_STOP,
	FIREBALL_ACTION_MAX
};
unsigned char __fastcall__  action_fireball(Chara* b) {
	unsigned char action = b->status & FIREBALL_ACTION_MASK;

	switch (action) {
	case FIREBALL_ACTION_INIT:
		action = FIREBALL_ACTION_MOVE | STATUS_BUSY;

	case FIREBALL_ACTION_MOVE:
		{
			// ˏoێ܂܈ړI
			unsigned char mcp;
			unsigned char id;

			b->x += b->dx;
			b->y += b->dy;
			mcp = map_getmcp(b->x, b->y);
			id = GET_CHIP_ID(mcp);

			if (id != MCP_FIELD) {
				nsd_play_se(SE_BREAK);
				action = FIREBALL_ACTION_STOP;

				if (id == MCP_FBLOCK) {
					map_set(b->x, b->y, MCP_BLOCK);
				}

			} else {
				int dx, dy;

				dx = chara[0].x - b->x;
				dy = chara[0].y - b->y;

				if (-6 <= dx && dx <= 6 && -6 <= dy && dy <= 6) {
					nsd_play_se(SE_BREAK);
					action = FIREBALL_ACTION_STOP;
					chara[0].dx += b->dx << 1;
					chara[0].dy += b->dy << 1;

				} else {

					dx = chara[1].x - b->x;
					dy = chara[1].y - b->y;

					if (-6 <= dx && dx <= 6 && -6 <= dy && dy <= 6) {
						nsd_play_se(SE_BREAK);
						action = FIREBALL_ACTION_STOP;
						chara[1].dx += b->dx << 1;
						chara[1].dy += b->dy << 1;
					}
				}
			}
		}
		break;

	case FIREBALL_ACTION_STOP:
		b->chr = CHR_NOTHING;
		b->status &= ~STATUS_BUSY;
		action = FIREBALL_ACTION_INIT;
		break;
	}

	b->status = (b->status & ~FIREBALL_ACTION_MASK) | action;

	return b->status;
}
